from typing import List

from sudoku_variants.rule.interface import Rule, WithData
from sudoku_variants.rule.jigsaw import Jigsaw
from sudoku_variants.rule.orthogonal import Orthogonal
from sudoku_variants.rule.sub_board import SubBoard

# because Python doesn't have extension


def to_name(rules: List[Rule]):
    return ", ".join(str(r) for r in rules)


def check_move(rules: List[Rule], board: List[List[int]], row: int, col: int, digit: int) -> bool:
    return all(rule.check_move(board, row, col, digit) for rule in rules)


def remove_candidates(
    rules: List[Rule], candidates: List[List[List[bool]]], row: int, col: int, digit: int
) -> List[List[List[bool]]]:
    for rule in rules:
        candidates = rule.remove_candidates(candidates, row, col, digit)
    return candidates


def populate_initial_data(rules: List[Rule]):
    for rule in rules:
        if isinstance(rule, WithData):
            rule.populate_initial_data()


def extract_data_from_board(rules: List[Rule], board: List[List[int]]):
    for rule in rules:
        if isinstance(rule, WithData):
            rule.extract_data_from_board(board)


def with_standard_rules(rules: List[Rule]) -> List[Rule]:
    standard_rules = [Orthogonal(), SubBoard()]
    for rule in rules:
        if isinstance(rule, Jigsaw):
            standard_rules = [Orthogonal()]

    return standard_rules + rules



from sudoku_variants.rule.interface import Rule, WithData
from sudoku_variants.rule.orthogonal import Orthogonal
from sudoku_variants.rule.sub_board import SubBoard
from sudoku_variants.rule.knight import Knight
from sudoku_variants.rule.king import King
from sudoku_variants.rule.consecutive import Consecutive
from sudoku_variants.rule.jigsaw import Jigsaw



import random

from typing import List, Tuple, Optional

from sudoku_variants.rule.interface import Rule, WithData
from sudoku_variants.sudoku_const import DIGITS, NUM_COL, NUM_ROW

ConsecutiveDataType = List[Tuple[Tuple[int, int], Tuple[int, int]]]


class Consecutive(Rule, WithData):
    offsets = [
        (-1, 0),
        (1, 0),
        (0, -1),
        (0, 1),
    ]

    def __init__(self, data: Optional[ConsecutiveDataType] = None) -> None:
        """
        data: a list of neighbours which is a pair of coordinates
        """
        super().__init__()
        self.data: ConsecutiveDataType
        if data is None:
            self.data = []
        else:
            self.data = data

    def check_move(self, board: List[List[int]], row: int, col: int, digit: int) -> bool:
        if digit not in DIGITS:
            return True

        cur_coord = (row, col)
        for offset in self.offsets:
            neighbour_coord = (row + offset[0], col + offset[1])
            pair1 = (cur_coord, neighbour_coord)
            pair2 = (neighbour_coord, cur_coord)

            if pair1 in self.data or pair2 in self.data:
                neighbour_digit = board[neighbour_coord[0]][neighbour_coord[1]]
                if neighbour_digit in DIGITS and abs(digit - neighbour_digit) != 1:
                    return False

        return True

    def populate_initial_data(self):
        self.data = []

    def extract_data_from_board(self, board: List[List[int]]):
        data = []
        for i in range(NUM_ROW):
            for j in range(NUM_COL):
                cur_coord = (i, j)
                cur_digit = board[i][j]

                for offset in self.offsets:
                    if (0 <= i + offset[0] < len(board)) and (0 <= j + offset[1] < len(board[0])):
                        neighbour_coord = (i + offset[0], j + offset[1])
                        neighbour_digit = board[neighbour_coord[0]][neighbour_coord[1]]
                        if abs(cur_digit - neighbour_digit) == 1:
                            data.append((cur_coord, neighbour_coord))

        size = random.randint(0, len(data))
        self.data = random.sample(data, size)



from typing import List

from sudoku_variants.rule.interface import Rule
from sudoku_variants.sudoku_const import DIGITS


class Knight(Rule):
    knight_moves = (
        (-2, 1),
        (-1, 2),
        (1, 2),
        (2, 1),
        (2, -1),
        (1, -2),
        (-1, -2),
        (-2, -1),
    )

    def check_move(self, board: List[List[int]], row: int, col: int, digit: int) -> bool:
        if digit not in DIGITS:
            return True

        for m in self.knight_moves:
            if (0 <= row + m[0] < len(board)) and (0 <= col + m[1] < len(board[0])):
                knight_digit = board[row + m[0]][col + m[1]]
                if digit == knight_digit:
                    return False

        return True



import random

from typing import List, Tuple, Optional, Dict, Set

from sudoku_variants.rule.interface import Rule, WithData
from sudoku_variants.sudoku_const import DIGITS, NUM_COL, NUM_ROW

JigsawDataType = List[List[int]]
JigsawPartitionType = Dict[int, List[Tuple[int, int]]]


def _default_jigsaw_layout():
    return [[row // 3 * 3 + col // 3 for col in range(NUM_COL)] for row in range(NUM_ROW)]


class Jigsaw(Rule, WithData):
    def __init__(self, data: Optional[JigsawDataType] = None) -> None:
        """
        data: a 2D board, where each cell indicates the index of its partition
        """
        super().__init__()
        self.data: JigsawDataType
        if data is None:
            self.data = _default_jigsaw_layout()
        else:
            self.data = data

        # to retrive all coords in a partition quicker
        self.partition: JigsawPartitionType = {}
        for i, row in enumerate(self.data):
            for j, partition in enumerate(row):
                self.partition.setdefault(partition, [])
                self.partition[partition].append((i, j))

    def check_move(self, board: List[List[int]], row: int, col: int, digit: int) -> bool:
        if digit not in DIGITS:
            return True

        partiton = self.data[row][col]
        coords_in_partition = self.partition[partiton]
        for coord in coords_in_partition:
            if (board[coord[0]][coord[1]] == digit) and (coord != (row, col)):
                return False

        return True

    def populate_initial_data(self, seed: Optional[int] = None):
        # reference: https://softwareengineering.stackexchange.com/questions/419048/algorithm-to-split-a-grid-of-squares-into-equally-sized-sections
        random.seed(seed)

        data = _default_jigsaw_layout()
        offsets = [(-1, 0), (1, 0), (0, -1), (0, 1)]

        def init_cell() -> Tuple[Tuple[int, int], int]:
            row = random.randint(0, len(data) - 1)
            col = random.randint(0, len(data[0]) - 1)
            coord = (row, col)
            partition = data[row][col]
            return (coord, partition)

        def is_partition_connected(partition: int) -> bool:
            coords: List[Tuple[int, int]] = []
            for i, row in enumerate(data):
                for j, p in enumerate(row):
                    if p == partition:
                        coords.append((i, j))
            size = len(coords)

            component: List[Tuple[int, int]] = [coords.pop()]
            has_new_coord = True
            while has_new_coord:
                has_new_coord = False
                new_coords: Set[Tuple[int, int]] = set()
                for coord in coords:
                    for offset in offsets:
                        neighbour = (coord[0] + offset[0], coord[1] + offset[1])
                        if neighbour in component:
                            new_coords.add(coord)

                if new_coords:
                    has_new_coord = True
                    component += new_coords
                    for c in new_coords:
                        coords.remove(c)

            return len(component) == size

        max_attempts = 300
        for i in range(max_attempts):
            coord1, partition1 = init_cell()
            coord2, partition2 = init_cell()
            data[coord1[0]][coord1[1]] = partition2
            data[coord2[0]][coord2[1]] = partition1

            is_data_connected = (
                (partition1 != partition2) and is_partition_connected(partition1) and is_partition_connected(partition2)
            )

            if not is_data_connected:
                data[coord1[0]][coord1[1]] = partition1
                data[coord2[0]][coord2[1]] = partition2

        self.data = data
        self.partition = {}
        for i, row in enumerate(self.data):
            for j, partition in enumerate(row):
                self.partition.setdefault(partition, [])
                self.partition[partition].append((i, j))

    def extract_data_from_board(self, board: List[List[int]]):
        pass



from typing import List


class Rule:
    def check_move(self, board: List[List[int]], row: int, col: int, digit: int) -> bool:
        """
        Assume the given board is valid, check if the new move is valid
        """
        raise NotImplementedError("Please implement this method")

    def remove_candidates(
        self, candidates: List[List[List[bool]]], row: int, col: int, digit: int
    ) -> List[List[List[bool]]]:
        """
        Remove candidates if the move is carried out to accelerate generation process
        NOTE: do not remove the candidates at (row, col)
        """
        return candidates

    def __str__(self) -> str:
        return type(self).__name__


class WithData:
    def populate_initial_data(self):
        """
        Used to reset data or initialize data before solving an empty sudoku
        """
        raise NotImplementedError("Please implement this method")

    def extract_data_from_board(self, board: List[List[int]]):
        """
        Used after solving sudoku and extract data from it
        """
        raise NotImplementedError("Please implement this method")



from typing import List

from sudoku_variants.rule.interface import Rule
from sudoku_variants.sudoku_const import DIGITS


class SubBoard(Rule):
    def check_move(self, board: List[List[int]], row: int, col: int, digit: int) -> bool:
        if digit not in DIGITS:
            return True

        sub_board_index = row // 3 * 3 + col // 3
        row_start = sub_board_index // 3 * 3
        row_end = row_start + 3
        col_start = sub_board_index % 3 * 3
        col_end = col_start + 3

        for i in range(row_start, row_end):
            for j in range(col_start, col_end):
                cur_digit = board[i][j]
                if (cur_digit == digit) and ((i != row) or (j != col)):
                    return False

        return True

    def remove_candidates(
        self, candidates: List[List[List[bool]]], row: int, col: int, digit: int
    ) -> List[List[List[bool]]]:
        if digit in DIGITS:
            sub_board_index = row // 3 * 3 + col // 3
            row_start = sub_board_index // 3 * 3
            row_end = row_start + 3
            col_start = sub_board_index % 3 * 3
            col_end = col_start + 3

            for i in range(row_start, row_end):
                for j in range(col_start, col_end):
                    if (i, j) != (row, col):
                        candidates[i][j][digit - 1] = False

        return candidates



from typing import List

from sudoku_variants.rule.interface import Rule
from sudoku_variants.sudoku_const import DIGITS, NUM_COL, NUM_ROW


class Orthogonal(Rule):
    def check_move(self, board: List[List[int]], row: int, col: int, digit: int) -> bool:
        if digit not in DIGITS:
            return True

        for col_index in range(NUM_COL):
            cur_digit = board[row][col_index]
            if (cur_digit == digit) and (col_index != col):
                return False

        for row_index in range(NUM_ROW):
            cur_digit = board[row_index][col]
            if (cur_digit == digit) and (row_index != row):
                return False

        return True

    def remove_candidates(
        self, candidates: List[List[List[bool]]], row: int, col: int, digit: int
    ) -> List[List[List[bool]]]:
        if digit in DIGITS:
            for col_index in range(NUM_COL):
                if col_index != col:
                    candidates[row][col_index][digit - 1] = False

            for row_index in range(NUM_ROW):
                if row_index != row:
                    candidates[row_index][col][digit - 1] = False

            candidates[row][col] = [False for _ in range(len(DIGITS))]

        return candidates



from typing import List, TypeVar, Tuple


T = TypeVar("T")


def shape_of_board(board: List[List]) -> Tuple[int, int]:
    return len(board), len(board[0])


def copy_board(board: List[List[T]]) -> List[List[T]]:
    return [[d for d in row] for row in board]
